import javax.sound.sampled.*;
import java.io.*;

/** implementace SoundBuffer zapisujici zvuk do souboru (format Wav)*/
public class SoundBufferFile extends SoundBuffer {

 private AudioFormat audioF;
 private float bps;//bytu/sec
 private RandomAccessFile rf;
 
 /** Konstruktor, vytvori buffer s vystupem do souboru wavfile a danym AudioFormatem */
 public SoundBufferFile(AudioFormat aud,String wavfile) throws IOException, FileNotFoundException {
  audioF=aud;
  bps=audioF.getChannels()*audioF.getFrameRate()*audioF.getFrameSize();
  File f=new File(wavfile);
  f.createNewFile();
  rf=new RandomAccessFile(f,"rw");
  rf.writeBytes("RIFF");
  rf.writeInt(-8);//filesize-8, filepos=start+4
  rf.writeBytes("WAVE");
  rf.writeBytes("fmt ");
  rf.writeInt(Util.swab((int)16));
  rf.writeShort(Util.swab((short)1));
  rf.writeShort(Util.swab((short)(audioF.getChannels())));
  rf.writeInt(Util.swab((int)audioF.getSampleRate()));
  rf.writeInt(Util.swab((int)bps));
  rf.writeShort(Util.swab((short)(audioF.getSampleSizeInBits()/8*audioF.getChannels())));
  rf.writeShort(Util.swab((short)(audioF.getSampleSizeInBits())));
  rf.writeBytes("data");
  rf.writeInt(-45);//filesize-45, filepos=start+40
 }
  
 /** Vrati AudioInputStream pro dany soubor f */
 public static AudioInputStream audioFromFile(File f) {
  try {
   AudioInputStream audioIn=AudioSystem.getAudioInputStream(f);
   return audioIn;
  } catch (Exception e) {
   return null;
  }
 }

 /** Prida msec milisekund ticha na vystupni stream */
 public boolean addSilence(int msec) {
  int tsize=(int)(bps*msec)/1000;
  byte[] buf = new byte[tsize];
  try {
   rf.write(buf,0,tsize);
  } catch (IOException e) {
   return false;
  }
  return true; 
 }

 /** Prida obsah AudioInputStreamu na vystupni stream */
 public boolean addStream(AudioInputStream in) throws IOException{
  int nRead = 0;
  byte[] buf = new byte[4096];
  while (nRead!=-1) {
   nRead=in.read(buf,0,buf.length);
   if (nRead>=0) {
    try {
     rf.write(buf,0,nRead);
    } catch (IOException e) {
     return false;
    }
   }
  }
  return true;
 }

 /** Prida obsah souboru na vystupni stream */
 public boolean addFile(File f) throws IOException{
  return addStream(audioFromFile(f));
 }

 /** Pocka na vyprazdneni sound bufferu - v teto implementaci dummy funkce */
 public void waitStop() {
  //dummy
 }

 /** Zavre soubor */
 public void close() throws SpeechException {
  try {
   int fsize=(int)rf.getFilePointer();
   rf.seek(4);
   rf.writeInt(Util.swab(fsize-8));
   rf.seek(40);
   rf.writeInt(Util.swab(fsize-45));
   rf.close();
  } catch (IOException e) {
   throw new SpeechException(e.toString());
  }
 }
}