import java.util.*;
import java.io.*;
import javax.sound.sampled.*;

/** synteza mluvene reci - trida na preprocesing a proneseni dane vety ci slova */
public class Speech {

 //** pauza mezi slovy (milisekund) */
 private final int PAUSE_MSEC=90;
 //** pauza za teckou (milisekund) */
 private final int PAUSE_DOT_MSEC=180;

 /** slovnik na preklad */
 private Vocabulary vocabulary=null;
 private boolean useVocabulary=true;

 /** cisla 0..19 */
 public static final String[] base={
 "nula",
 "jedna",
 "dva",
 "ti",
 "tyi",
 "pt",
 "est",
 "sedm",
 "osm",
 "devt",
 "deset",
 "jedenct",
 "dvanct",
 "tinct",
 "trnct",
 "patnct",
 "estnct",
 "semdnct",
 "osmnct",
 "devatenct"
 };
 /** cisla desitek 20 .. 90 */
 public static final String[] base10={
 "","",//0,10
 "dvacet",
 "ticet",
 "tyicet",
 "padest",
 "edest",
 "sedmdest",
 "osmdest",
 "devadest",
 };
 /** cisla stovek do 500 */
 public static final String[] base100={
 "",
 "sto",
 "dvst",
 "tista",
 "tyista",
 };
 /** nazvy vyssich radu */
 public static final String[] base1K={
 "tisc", //1		10^3
 "tisce",//2,3,4
 "tisc", //5+
 "milin", //1		10^6
 "miliny",//2,3,4
 "milin",//5+
 "miliarda",//1		10^9
 "miliardy",//2,3,4
 "miliard", //5+
 "bilin", //1		10^12
 "biliny",//2,3,4
 "bilin",//5+
 "trilin", //1		10^18
 "triliny",//2,3,4
 "trilin",//5+
 "kvadrilin", //1	10^24
 "kvadriliny",//2,3,4
 "kvadrilin",//5+
 };

 /** vrati slovni vyjadreni vyssi casti cislic s vyssim radem */
 private static String highNumberToString(long n,int off) {
  if (n==1) return base1K[off*3];
  if (n<=4) return numberToString(n)+base1K[off*3+1];
  return numberToString(n)+base1K[off*3+2];
 }

 /** prevede cislici na jeji slovni vyjadreni. Dokaze prevest vse co se vejde do typu long */
 public static String numberToString(long n) {
  int i;
  long baseNum;
  if (n<0) return "mnus "+numberToString(-n);
  if (n<=19) return base[(int)n];
  if (n<100) {
   int rem=(int)n%10;
   return base10[(int)n/10]+((rem>0)?numberToString(rem):"");
  }
  if (n<500) {
   int rem=(int)n%100;
   return base100[(int)n/100]+((rem>0)?numberToString(rem):"");
  }
  if (n<2000) {
   int rem=(int)n%100;
   if (((int)n/100)!=10) //vyjimka: desetset je tisic
    return base[(int)n/100]+"set"+((rem>0)?numberToString(rem):"");
  }
  baseNum=1000;
  for (i=0;i<3;i++) {//3,6,9 - vyssi rady
   if (n<1000*baseNum) {
    long rem=n%baseNum;
    return highNumberToString(n/baseNum,i)+((rem>0)?numberToString(rem):"");
   }
   baseNum*=1000;
  }
  for (i=0;i<3;i++) {//12,18,24 - jeste vyssi rady
   if (n/1000000<baseNum) {
    long rem=n%baseNum;
    return highNumberToString(n/baseNum,i+3)+((rem>0)?numberToString(rem):"");
   }
   baseNum*=1000000;
  }
  return "mnoho";//sem to stejne nikdy nedojde  
 }

 /** vraci true pokud je znak cislo */
 public static boolean isNumber(char z) {
  return (z>='0' && z<='9');
 }

 /** Preprocessing daneho textu (nahrazeni cislic a symbolu slovy a tak...) */
 private static String preprocess(String s) {
  int i;
  char curr;					//aktualni znak
  char prev=' ';				//predesly znak
  StringBuffer tmp=new StringBuffer();
  s=s.toLowerCase();//na mala pismena
  //faze 1 - nahrazeni cislic a symbol slovy
  for(i=0;i<s.length();i++) {
   curr=s.charAt(i);
   if (curr==' ' && prev==' ') continue;	//vice mezer za sebou ignorovat
   if (isNumber(curr)) {		//cislice
    long num=0;
    num=(curr-'0');
    while(i+1<s.length() && isNumber(s.charAt(i+1))) {
     i++;
     num*=10;
     num+=(s.charAt(i)-'0');
    }
    if (prev=='-') {//minus
     tmp.setLength(tmp.length()-1);//odmazne minus
     num=-num;
    }
    tmp.append(numberToString(num));
   } else if (curr=='@') {
    tmp.append(" zavin ");
   } else if (curr=='+') {
    tmp.append(" plus ");
   } else if (curr=='*') {
    tmp.append(" krt ");
   } else if (curr=='/') {
    tmp.append(" lomeno ");
   } else if (curr=='=') {
    tmp.append(" rovn se ");
   } else if (curr=='^') {
    tmp.append(" na ");
   } else if (curr=='%') {
    tmp.append(" procent ");
   } else tmp.append(curr);	//as-is
   prev=curr;
  }
  return tmp.toString();
 }

 /** Preprocessing faze 2 -> zmeny ve vyslovnosti */
 private static String convert(String tmp) {
  char curr;					//aktualni znak
  char prev=' ';				//predesly znak
  StringBuffer out=new StringBuffer();
  //faze 2 - zmena vyslovnosti (cestina ma precejen par vyjimek)
  for(int i=0;i<tmp.length();i++) {//cislice
   curr=tmp.charAt(i);
   if (curr=='x') {				//X se vyslovi jako KS
    out.append("ks");
   } else if (curr=='') {			//E s hackem
    if (prev=='m') out.append("e");//ME~ -> MN~E
    else if (prev=='t') {	//TE~ -> T~E
     out.setCharAt(out.length()-1,'');out.append("e");
    } else if (prev=='d') {	//DE~ -> D~E
     out.setCharAt(out.length()-1,'');out.append("e");
    } else if (prev=='n') {	//NE~ -> N~E
     out.setCharAt(out.length()-1,'');out.append("e");
    } else out.append("je");	//E~ -> JE
   } else if (curr=='') {			// =  (krouzek nebo carka jsou stejny u U co do vyslovnosti)
     out.append("");
   } else if (prev=='c' && curr=='h') {		//Ch je jedna hlaska, nahrad c co je v bufferu. Ch si ted oznacime jako X
     out.setCharAt(out.length()-1,'x');
   } else if (curr=='w') {			//W -> V
     out.append("v");    
   } else if (curr=='y') {			//Y -> I
     out.append("i");    
   } else if (curr=='') {			//Y -> I s carkou
     out.append("");    
   } else out.append(curr);			//as-is
   prev=curr;
  }
  return out.toString();
 }

 /** HashMap asociujici k difonu dany zvuk */
 private HashMap sndFiles;
 /** format audio souboru - vsechny musi mit stejny a to tento, ktery se vezme z prvniho nacteneho souboru */
 private AudioFormat allFormat;

 /** inicializace - adresar s wavy */
 private void init(String waveDir) throws SpeechException {
  try {
   File mapping=new File(waveDir,"sound.mapping");
   Properties p=new Properties();
   p.load(new FileInputStream(mapping));
   String key,name;
   String format_s=null;
   sndFiles=new HashMap();
   for (Enumeration wfiles=p.propertyNames();wfiles.hasMoreElements();) {
    key=(String)wfiles.nextElement();
    name=p.getProperty(key);
    if (key.equals("-")) key=" ";
    File snd=new File(waveDir,name);
    if (snd==null) throw new SpeechException("Sound file '"+name+"' for '"+key+"' is missing");
    AudioInputStream audioIn=SoundBuffer.audioFromFile(snd);
    if (audioIn==null) throw new SpeechException("Sound file '"+name+"' for '"+key+"' cannot be loaded");
    String format = audioIn.getFormat().toString();
    if (format_s!=null) {
     if (!format_s.equals(format)) {
      sndFiles.clear();
      throw new SpeechException("All sounds must be of the same format");
     }
    } else {
     format_s=format;
     allFormat=audioIn.getFormat();
    }
    sndFiles.put(key,snd);
   }
  } catch (IOException e) {
   throw new SpeechException("Can't load sound mapping files");
  }
 }

 /** defaultni konstruktor */
 public Speech() throws SpeechException {
  init("sound");
 }

 /** konstruktor s jinym datovym adresarem (wavy) */
 public Speech(String datadir) throws SpeechException {
  init(datadir);
 }

 /** snazi se pronest dana pismenka presne jednim fonemem */
 private boolean trySayExact(SoundBuffer snd,String s) {
  if (s.equals(" ") || s.equals("\n")) {
   snd.addSilence(PAUSE_MSEC);
//   System.out.print(" ");
   return true;
  }
  if (s.matches("^[\\.,!\\?]$")) {
   snd.addSilence(PAUSE_DOT_MSEC);
//   System.out.print(".");
   return true;
  }
  File f=(File)sndFiles.get(s);
  try {
   if (f!=null) {
//    System.out.print("("+s+")");
    snd.addFile(f);
    return true;
   }
  } catch (Exception e){}//skonci na return false
  return false;
 }

 /** Pismena jez jsou samohlaskami */
 private static final String samohlasky="aeiou";

 /** snazi se pronest co nejvice pismenek ze slova jednim fonemem. Vraci pocet pismenek ktere pronesl */
 private int trySay(SoundBuffer snd,String s) {
//  System.out.println("trysay:"+s);
  int vlen=Math.min(3,s.length());//delka segmentu ktery zkusime rict
  if (vlen==3) {
   if (/*samohlasky.indexOf(s.charAt(0))==-1
   &&  */samohlasky.indexOf(s.charAt(1))==-1
   &&  samohlasky.indexOf(s.charAt(2))!=-1
   ) { //2 souhlasky a samohlaska .... lepsi ric m-no nez mn-o
       //to same samo-sou-samo hlasky
    if (trySayExact(snd,s.substring(0,1))) return 1;
    return 1; //?? nevyslovitelna hlaska
   }
  }
  for(int i=vlen;i>0;i--) {
   if (trySayExact(snd,s.substring(0,i))) return i;
  }
//  System.out.print("[!"+s.substring(0,1)+"!]");
  return 1;//kdyz nevim o co de, tak holt asi preskocim ...
 }

 /** Pronese danou vetu nebo text, vystupem je defaulni zvukova karta, funkce ceka na dokonceni */
 public void say(String s) throws IOException, LineUnavailableException, SpeechException {
  SoundBuffer snd=new SoundBufferCard(allFormat);
  say(s,snd);
 }

 /** Pronese danou vetu nebo text, vystupem je Wav soubor filename, funkce ceka na dokonceni */
 public void sayToFile(String s,String filename) throws IOException, LineUnavailableException, SpeechException {
  SoundBuffer snd=new SoundBufferFile(allFormat,filename);
  say(s,snd);
 }

 /** Pronese danou vetu nebo text, vystupem je SoundBuffer snd, funkce ceka na dokonceni */
 public void say(String s,SoundBuffer snd) throws IOException, SpeechException {
  if (useVocabulary && vocabulary!=null) s=vocabulary.processSentence(s);
  s=preprocess(s);
  s=convert(s);
  //System.out.println(s);
  int n=0;
  //Vzit to po slovech greedy algoritmem, soupat do vystupu
  while(n<s.length()) {
   int end=n+3;
   if (end>s.length()) end=s.length();
   String key=s.substring(n,end);
   n+=trySay(snd,key);
  }
  snd.waitStop();
  snd.close();
 }

 /** nahraje slovnik na ohackovani nehackovanych vyrazu */
 public void loadVocabulary(String filename,String encoding) throws Exception  {
  vocabulary=new Vocabulary(filename,encoding);
 }

 /** nahraje slovnik na ohackovani nehackovanych vyrazu */
 public void loadVocabulary(String filename) throws Exception {
  vocabulary=new Vocabulary(filename);
 }

 /** nastavi jestli se bude pouzivt slovnik k ohackovani ceskych slov */ 
 public void useVocabulary(boolean use) {
  useVocabulary=use;
 }

 /** vrati aktualne pouzity slovnik (metoda hlavne pro debugovaci ucely) */
 public Vocabulary getVocabulary() {
  return vocabulary;
 }

}
